/*
 * cache.c
 *
 * Copyright (C) 1994, 1995, John D. Kilburg (john@cs.unlv.edu)
 *
 * See copyright.h for details.
 */
#include "copyright.h"

#include <stdio.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <sys/stat.h>
#include <sys/param.h>

#if defined(SYSV) || defined(SVR4)|| defined(__arm)
#include <dirent.h>
#define DIRSTUFF struct dirent
#else
#include <sys/dir.h>
#define DIRSTUFF struct direct
#endif

#ifndef MAXPATHLEN
#define MAXPATHLEN 1024
#endif

#include "md5.h"
#include "common.h"
#include "wwwio.h"
#include "cache.h"

static size_t cache_size = 0;
static size_t cache_max_size = 0;
static time_t cache_ttl = 0;
static char *cache_dir = NULL;
static CacheEntry *clist = NULL;
static char *cache_domains = NULL;

static CacheEntry *FindCacheEntry _ArgProto((char *));
static int HasExpired _ArgProto((time_t, time_t));
static void InsertCacheEntry _ArgProto((char *, time_t, time_t,
					off_t, time_t));
static void DeleteCacheEntry _ArgProto((CacheEntry *));
static char *urlToCacheName _ArgProto((URLParts *));
static void ScanCacheDirectory _ArgProto((void));
static int FlushCache _ArgProto((void));

/*
 * HasExpired
 */
static int
HasExpired(expires, mtime)
time_t expires, mtime;
{
  time_t now;

  now = time(NULL);
  if ((expires > 0 && now > expires) ||
      (now - mtime > cache_ttl && cache_ttl > 0))
  {
    return(1);
  }
  return(0);
}

/*
 * FindCacheEntry
 */
static CacheEntry *
FindCacheEntry(filename)
char *filename;
{
  CacheEntry *c;

  for (c = clist; c; c = c->next)
  {
    if (strcmp(filename, c->path) == 0) return(c);
  }

  return(NULL);
}

/*
 * InsertCacheEntry
 */
static void
InsertCacheEntry(path, atime, mtime, size, expires)
char *path;
time_t atime;
time_t mtime;
off_t size;
time_t expires;
{
  CacheEntry *n, *c, *t;

  n = (CacheEntry *)alloc_mem(sizeof(CacheEntry));
  n->path = alloc_string(path);
  n->atime = atime;
  n->mtime = mtime;
  n->size = size;
  n->prev = NULL;
  n->next = NULL;
  n->expires = expires;

  t = NULL;
  c = clist;
  while (c)
  {
    if (c->atime > atime)
    {
      if (t != NULL) t->next = n;
      else clist = n;
      n->next = c;
      n->prev = t;
      return;
    }
 
    t = c;
    c = c->next;
  }

  if (t != NULL) t->next = n;
  else clist = n;
  n->prev = t;

  return;
}

/*
 * DeleteCacheEntry
 */
static void
DeleteCacheEntry(c)
CacheEntry *c;
{
  unlink(c->path);
  cache_size -= c->size;

  if (c->prev == NULL) clist = c->next;
  else c->prev->next = c->next;

  if (c->next != NULL) c->next->prev = c->prev;

  free_mem(c->path);
  free_mem((char *)c);

  return;
}

/*
 * ScanCacheDirectory
 */
static void
ScanCacheDirectory()
{
  FILE *fp;
  char buffer[BUFSIZ];
  char path[MAXPATHLEN + 1];
  DIR *dp;
  struct stat s;
  time_t expires = 0;
  DIRSTUFF *de;
  URLParts *up;
  char *value;

  /*
   * Scan the cache directory for cache files.
   */
  dp = opendir(cache_dir);
  if (dp != NULL)
  {
    while ((de = readdir(dp)) != NULL)
    {
      /*
       * Is the prefix right?
       */
      if (strncmp("ccf", de->d_name, 3) == 0)
      {
	/*
	 * Get the full path of the cache file.
	 */
	sprintf (path, "%s/%s", cache_dir, de->d_name);
	if (stat(path, &s) != 0) continue;
	
	/*
	 * Scan the header and look for 'Expires' and 'X-URL'.
	 */
	up = NULL;
	expires = 0;
	fp = fopen(path, "r");
	if (fp != NULL)
	{
	  Document *d = CreateDocument();
	  
	  while (fgets(buffer, sizeof(buffer), fp) != NULL)
	  {
	    if (buffer[0] == '\n') break;
	    
	    MIMEParseHeaderLine(d, buffer);
	  }
	  fclose(fp);
	  
	  if ((value = MIMEGetFieldValue(d->mflist, "expires", 0)) != NULL)
	  {
	    expires = ParseExpiresDate(value);
	  }
	  if ((value = MIMEGetFieldValue(d->mflist, "X-URL", 0)) != NULL)
	  {
	    up = ParseURL(value);
	  }
	  
	  DestroyDocument(d);
	  
	  /*
	   * If the document has not expired then add it to the list
	   * otherwise remove the file.
	   */
	  if (HasExpired(expires, s.st_mtime) == 0)
	  {
	    if (FindCacheEntry(path) == NULL)
	    {
	      InsertCacheEntry(path, s.st_atime, s.st_mtime,
			       s.st_size, expires);
	      cache_size += s.st_size;
	    }
	  }
	  else unlink(path);
	}
      }
    }    
    closedir(dp);
  }
  
  return;
}

/*
 * CacheInit
 */
void
CacheInit(dir, max_size, ttl, domains)
char *dir;
int max_size;
int ttl;
char *domains;
{
  cache_dir = alloc_string(dir);
  cache_max_size = (off_t)max_size;
  cache_ttl = ttl;
  if (domains != NULL) cache_domains = alloc_string(domains);
  else cache_domains = NULL;

  ScanCacheDirectory();

  return;
}

/*
 * urlToCacheName
 *
 * uses MD5 to make a name for the cache.
 */
static char *
urlToCacheName(up)
URLParts *up;
{
  static char name[MAXPATHLEN + 1];
  struct MD5Context context;
  unsigned char digest[16];
  unsigned char hex[33];
  unsigned char *url;
  unsigned char *cp, *ep, *dp;

  url = (unsigned char *)MakeURL(up, 0);

  MD5Init(&context);
  MD5Update(&context, url, strlen(url));
  MD5Final(digest, &context);
  free_mem(url);

  cp = hex;
  for (dp = digest, ep = digest + (sizeof(digest) / sizeof(digest[0]));
           dp < ep; cp += 2)
  {
    (void)sprintf(cp, "%02x", *dp++ & 0xff);
  }
  *cp = '\0';

  /* chop it off to have reasonable length file names */
  hex[11] = '\0';

  sprintf(name, "%s/ccf%s", cache_dir, hex);

  return(name);
}

/*
 * IsCached
 *
 * Returns 1 if the specified URL is in the cache, otherwise returns 0.
 */
int
IsCached(up)
URLParts *up;
{
  CacheEntry *c;
  char *filename;
  struct stat s;

  if (up->hostname == NULL) return(0);

  filename = urlToCacheName(up);

  if (stat(filename, &s) == -1) return(0);
  
  /*
   * If the document is too old then don't use it.
   */
  c = FindCacheEntry(filename);
  if (c == NULL)
  {
    unlink(filename);
    return(0);
  }

  if (HasExpired(c->expires, c->mtime) != 0)
  {
    DeleteCacheEntry(c);
    return(0);
  }

  return(1);
}

/*
 * FlushCache
 */
static int
FlushCache()
{
  if (clist != NULL)
  {
    DeleteCacheEntry(clist);
    return(1);
  }

  return(0);
}

/*
 * CacheWrite
 */
void
CacheWrite(d, up)
Document *d;
URLParts *up;
{
  FILE *fp;
  char *u;
  char *filename;
  off_t size = 0;
  time_t expires, now;
  char *value;
  MIMEField *mf;

  if (d->from_cache != 0) return;

  now = time(NULL);
  if ((value = MIMEGetFieldValue(d->mflist, "expires", 0)) != NULL)
  {
    expires = ParseExpiresDate(value);
    if (expires < now) return;
  }
  else expires = 0;

  if (d->cache == 0 || d->from_cache != 0 ||
      up == NULL || up->hostname == NULL)
  {
    return;
  }

  filename = urlToCacheName(up);
  
  size = (off_t)d->len;  
  while (cache_max_size > 0 &&
	 ((size + cache_size) > cache_max_size))
  {
    if (!FlushCache()) return;
  }
  
  fp = fopen(filename, "w");
  if (fp == NULL) return;

  u = MakeURL(up, 1);
  fprintf (fp, "X-URL: %s\n", u);
  free_mem(u);

  for (mf = d->mflist; mf != NULL; mf = mf->next)
  {
    fprintf (fp, "%s: %s\n", mf->name, mf->value);
  }

  fprintf (fp, "\n");

  if (fwrite(d->text, 1, d->len, fp) < d->len)
  {
    fclose(fp);
    unlink(filename);
    return;
  }

  fclose(fp);
  
  cache_size += size;
  
  InsertCacheEntry(filename, now, now, size, expires);
  
  return;
}

/*
 * CacheClean
 */
void
CacheClean(cleanup)
int cleanup;
{
  CacheEntry *c, *t;

  c = clist;
  while (c)
  {
    t = c;
    c = c->next;
    if (t->path != NULL)
    {
      if (cleanup != 0) unlink(t->path);
      free_mem(t->path);
    }
    free_mem((char *)t);
  }

  if (cache_dir != NULL) free_mem(cache_dir);
  if (cache_domains != NULL) free_mem(cache_domains);

  return;
}


/*
 * CacheRead
 */
Document *
CacheRead(up)
URLParts *up;
{
  char *filename;
  Document *d;
  FILE *fp;
  char buffer[BUFSIZ];
  CacheEntry *c;
  char *value;

  if (up->input_count > 0 || up->hostname == NULL) return(NULL);

  filename = urlToCacheName(up);

  fp = fopen(filename, "r");
  if (fp == NULL) return(NULL);
  
  /*
   * If the document is too old then don't use it.
   */
  c = FindCacheEntry(filename);
  if (c == NULL)
  {
    fclose(fp);
    unlink(filename);
    return(NULL);
  }

  if (HasExpired(c->expires, c->mtime))
  {
    fclose(fp);
    DeleteCacheEntry(c);
    return(NULL);
  }

  d = CreateDocument();

  /*
   * Read MIME header
   */
  while (fgets(buffer, sizeof(buffer), fp) != NULL)
  {
    if (buffer[0] == '\n') break;

    MIMEParseHeaderLine(d, buffer);
  }

  if ((value = MIMEGetFieldValue(d->mflist, "content-length", 0)) != NULL)
  {
    d->len = atoi(value);
  }

  /*
   * If content length is zero then flush
   */
  if (d->len == 0)
  {
    fclose(fp);
    DestroyDocument(d);
    DeleteCacheEntry(c);
    return(NULL);
  }

  /*
   * Read the contents.
   */
  d->text = (char *)alloc_mem(d->len + 1);
  if (fread(d->text, 1, d->len, fp) < d->len)
  {
    fclose(fp);
    DestroyDocument(d);
    DeleteCacheEntry(c);
    return(NULL);
  }
  d->text[d->len] = '\0';

  fclose(fp);

  d->from_cache = 1;

  return(d);
}
