/*
 * inline.c
 *
 * Copyright (c) 1995, John D. Kilburg (john@cs.unlv.edu)
 *
 * See copyright.h for more information.
 *
 */
#include "copyright.h"

#include <stdio.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include "common.h"
#include "WWWP.h"
#include "xloadimage.h"
#include "gif.h"
#include "dither.h"

#define ANCHOR_BOX_LINE_WIDTH 1

typedef struct _inline_state
{
  WWWWidget rw;                   /* always need this guy around */
  Document *d;                    /* the inline document */
  URLParts *up;                   /* the URL of the inline doc */
  XImage *xi;                     /* X Image */
  gifState *gs;                   /* GIF decoder state */
  ditherState *ds;                /* dither state */
  XColor bg;                      /* background color */
  RenderPart *next;               /* next image in same image list */
  RenderPart *mp;                 /* master image */
} InlineState;

/*
 *
 * Private Functions
 *
 */

static int HandleInlineDocument _ArgProto((Document *, void *, int));
static RenderPart *CreateInlineState _ArgProto((WWWWidget, MLTag *,
						URLParts *));
static void GetImageDimensions _ArgProto((InlineState *, int *, int *));
static void AddImageRenderPart _ArgProto((WWWWidget, RenderPart *, int, int));
static void DestroyInlineState _ArgProto((InlineState *));
static void ImageToXImage _ArgProto((RenderPart *, int));
static void RenderInline _ArgProto((WWWWidget, RenderPart *, XEvent *));
static void SetupInline _ArgProto((WWWWidget, RenderPart *, Region, int, int));
static void AddFakeRenderPart _ArgProto((WWWWidget));
static Boolean ProcessInlineDocument _ArgProto((RenderPart *));

/*
 * SetupInline
 */
static void
SetupInline(rw, rp, r, line_y, line_height)
WWWWidget rw;
RenderPart *rp;
Region r;
int line_y;
int line_height;
{
  rp->y = line_y + line_height - rp->height;

  if (XRectInRegion(r, rp->x, rp->y, rp->width, rp->height))
  {
    RenderInline(rw, rp, NULL);
  }

  return;
}

/*
 * RenderInline
 *
 * Renders an inline image
 */
static void
RenderInline(rw, rp, xe)
WWWWidget rw;
RenderPart *rp;
XEvent *xe;
{
  InlineState *ci = (InlineState *)GetRenderState(rp);
  Display *dpy = rw->www.dpy;
  Window win = XtWindow(rw->www.child);
  GC gc = rw->www.gc;
  XImage *xi;
  int offset;
  int srcx, srcy;
  int dstx, dsty;
  int x, y;
  int bx, by;
  unsigned int width, height;

  if (ci->mp != NULL) xi = ((InlineState *)GetRenderState(ci->mp))->xi;
  else xi = ci->xi;

  if (xi == NULL) return;

  if (rp->anchorPart != NULL)
  {
    XSetForeground(dpy, gc, rw->www.anchorColor);
    XDrawRectangle(dpy, win, gc,
		   rp->x, rp->y,
		   rp->width, rp->height);
    XSetForeground(dpy, gc, rw->www.foreground);

    offset = ANCHOR_BOX_LINE_WIDTH;
  }
  else offset = 0;

  x = rp->x + offset;
  y = rp->y + offset;

  if (xe == NULL)
  {
    srcx = 0;
    srcy = 0;
    dstx = x;
    dsty = y;
    width = xi->width;
    height = xi->height;
  }
  else
  {
    dstx = x < xe->xexpose.x ? xe->xexpose.x:x;
    dsty = y < xe->xexpose.y ? xe->xexpose.y:y;

    srcx = dstx - x;
    srcy = dsty - y;

    bx = x + xi->width;
    if (xe->xexpose.x + xe->xexpose.width < bx)
    {
      bx = xe->xexpose.x + xe->xexpose.width;
    }
    width = bx - dstx;

    by = y + xi->height;
    if (xe->xexpose.y + xe->xexpose.height < by)
    {
      by = xe->xexpose.y + xe->xexpose.height;
    }
    height = by - dsty;
  }

  XPutImage(dpy, win, gc, xi,
	    srcx, srcy,
	    dstx, dsty,
	    width, height);
  
  return;
}

/*
 * GetImageDimensions
 */
static void
GetImageDimensions(ci, width, height)
InlineState *ci;
int *width, *height;
{
  if (ci->gs != NULL) gifDimensions(ci->gs, width, height);
  else
  {
    *width = 0;
    *height = 0;
  }

  return;
}

/*
 * DoColorInline
 */
static void
DoColorInline(ci, line)
InlineState *ci;
int line;
{
  int color_index;
  byte *pixptr;
  int x;
  Pixel pixel;
  Image *image = ci->gs->image;

  pixptr = image->data + (image->pixlen * (line * image->width));
  for (x = 0; x < image->width; x++)
  {
    color_index = memToVal(pixptr, image->pixlen);

    if (color_index == ci->gs->transparent)
    {
      pixel = ci->bg.pixel;
    }
    else
    {
      pixel = GetPixel(ci->rw,
		       image->rgb.red[color_index],
		       image->rgb.green[color_index],
		       image->rgb.blue[color_index]);
    }
    
    XPutPixel(ci->xi, x, line, pixel);
    pixptr += image->pixlen;
  }

  return;
}

/*
 * DoBitInline
 */
static void
DoBitInline(ci, line)
InlineState *ci;
int line;
{
  byte *pixptr;
  int x;
  Pixel pixel;
  Image *image = ci->ds->image;

  ditherAddLine(ci->ds, line);

  pixptr = image->data + line * image->width;
  for (x = 0; x < image->width; x++, pixptr++)
  {
    if (*pixptr == 0) pixel = ci->bg.pixel;
    else pixel = BlackPixel(ci->rw->www.dpy, ci->rw->www.screen);
    
    XPutPixel(ci->xi, x, line, pixel);
  }

  return;
}

/*
 * ImageToXImage
 */
static void
ImageToXImage(rp, line)
RenderPart *rp;
int line;
{
  InlineState *ci = (InlineState *)GetRenderState(rp);
  WWWWidget rw = ci->rw;
  Display *dpy = rw->www.dpy;
  Window win = XtWindow(rw->www.child);
  GC gc = rw->www.gc;
  InlineState *fci;
  int offset;
  RenderPart *tp;

  if (ci->xi == NULL)
  {
    Visual *v;
    int width, height;
    int depth;

    GetImageDimensions(ci, &width, &height);
    
    v = DefaultVisual(dpy, rw->www.screen);
    depth = DefaultDepth(dpy, rw->www.screen);

    ci->xi = XCreateImage(dpy, v,
			  depth,
			  ZPixmap, 0, NULL,
			  width, height,
			  8, 0);
    ci->xi->data = (char *)calloc_mem(1, height * ci->xi->bytes_per_line);
    XAddPixel(ci->xi, ci->bg.pixel);

    if (rw->www.ncolors < 16) ci->ds = ditherInit(ci->gs->image);

    if (!RenderPartAdded(rp)) AddImageRenderPart(rw, rp, width, height);
  }

  if (ci->ds != NULL) DoBitInline(ci, line);
  else DoColorInline(ci, line);

  for (tp = rp, fci = ci; tp != NULL;
       fci = (InlineState *)GetRenderState(tp), tp = fci->next)
  {
    if (tp->renderable)
    {
      if (tp->anchorPart != NULL) offset = ANCHOR_BOX_LINE_WIDTH;
      else offset = 0;

      XPutImage(dpy, win, gc, ci->xi,
		0, line,
		tp->x + offset,
		tp->y + line + offset,
		tp->width, 1);
    }
  }

  return;
}

/*
 * AddImageRenderPart
 */
static void
AddImageRenderPart(rw, rp, width, height)
WWWWidget rw;
RenderPart *rp;
int width, height;
{
  int xw;

  if (rp->anchorPart != NULL) xw = ANCHOR_BOX_LINE_WIDTH;
  else xw = 0;

  AddRenderPart(rw, rp, width + xw * 2, height + xw * 2, False);

  return;
}

/*
 * AddFakeRenderPart
 */
static void
AddFakeRenderPart(rw)
WWWWidget rw;
{
  RenderPart *rp;

  rp = CreateRenderPart(rw, NULL, NULL, NULL, NULL, NULL, 0);
  AddRenderPart(rw, rp, 0, 0, False);

  return;
}

/*
 * ProcessInlineDocument
 *
 * Returns False if the download should continue.
 */
static Boolean
ProcessInlineDocument(rp)
RenderPart *rp;
{
  InlineState *ci = (InlineState *)GetRenderState(rp);
  Document *d = ci->d;
  char *value;
  URLParts *up;

  if ((value = MIMEGetFieldValue(d->mflist, "location", 0)) != NULL)
  {
    if ((up = ParseURL(value)) != NULL)
    {
      DestroyDocument(d);

      DestroyURLParts(ci->up);
      ci->up = up;

      up->reload = ci->rw->www.reload ? 1:0;
      LoadDocument((IOContext)ci->rw, up, HandleInlineDocument, ci);

      return(True);
    }
  }

  if ((value = MIMEGetFieldValue(d->mflist, "content-type", 0)) != NULL)
  {
    if (strcasecmp(value, "image/gif") == 0)
    {
      ci->gs = gifInit(ImageToXImage, rp);
    }
    else AddFakeRenderPart(ci->rw);
  }
  else AddFakeRenderPart(ci->rw);

  return(False);
}

/*
 * HandleInlineDocument
 */
static int
HandleInlineDocument(d, closure, finished)
Document *d;
void *closure;
int finished;
{
  RenderPart *rp = (RenderPart *)closure;
  InlineState *ci = (InlineState *)GetRenderState(rp);

  ci->d = d;

  if (d->status == DS_NEEDS_AUTH)
  {
    if (ci->up->username == NULL && FetchAuthInfo(ci->rw, ci->d, ci->up))
    {
      DestroyDocument(ci->d);
      ci->d = NULL;
      LoadDocument((IOContext)(ci->rw), ci->up, HandleInlineDocument, rp);
      return(-1);
    }
  }

  if (d->status != DS_OK)
  {
    if (d->status != DS_ABORT && !RenderPartAdded(rp))
    {
      AddFakeRenderPart(ci->rw);
    }
    return(-1);
  }

  if (ci->gs == NULL)
  {
    if (ProcessInlineDocument(rp)) return(-1);
  }
  if (ci->gs != NULL && d->len > 0) gifAddData(ci->gs, d->text, d->len);

  if (finished != 0)
  {
    if (!RenderPartAdded(rp)) AddFakeRenderPart(ci->rw);
  }

  return(0);
}

static void
DestroyInlineState(ci)
InlineState *ci;
{
  if (ci->gs != NULL) gifDestroy(ci->gs);
  if (ci->up != NULL) DestroyURLParts(ci->up);
  if (ci->d != NULL) DestroyDocument(ci->d);
  if (ci->xi != NULL) XDestroyImage(ci->xi);
  if (ci->ds != NULL) ditherDestroy(ci->ds);

  return;
}

static RenderPart *
CreateInlineState(rw, p, up)
WWWWidget rw;
MLTag *p;
URLParts *up;
{
  InlineState *ci;
  RenderPart *rp;

  rp = CreateRenderPart(rw, p,
			SetupInline,
			RenderInline, DestroyInlineState, NULL,
			sizeof(InlineState));
  ci = (InlineState *)GetRenderState(rp);
  ci->rw = rw;
  ci->d = NULL;
  ci->up = up;
  ci->gs = NULL;
  ci->ds = NULL;
  ci->xi = NULL;
  ci->next = NULL;
  ci->mp = NULL;

  return(rp);
}

/*
 *
 * Public functions
 *
 */

/*
 * HandleImg
 */
void
HandleImg(rw, p, end)
WWWWidget rw;
MLTag *p;
int end;
{
  u_char *url;
  URLParts *aup;
  URLParts *lup;
  URLParts *bup;
  InlineState *ci;
  RenderPart *rp;
  RenderPart *inline_rp;
  int width, height;

  if (end != 0) return;

  if ((url = MLFindAttribute(p, "src")) == NULL) return;

  /*
   * Check if this image is already being downloaded.
   */
  for (rp = rw->www.rphead; rp != NULL; rp = rp->mnext)
  {
    if (rp->p != NULL && rp->p->id == TAG_IMG)
    {
      char *xurl;
      
      if ((xurl = MLFindAttribute(rp->p, "src")) != NULL &&
	  strcmp(xurl, url) == 0)
      {
	InlineState *xci;
	
	if ((xci = (InlineState *)GetRenderState(rp)) != NULL)
	{
	  inline_rp = CreateInlineState(rw, rp->p, NULL);
	  ci = (InlineState *)GetRenderState(inline_rp);

	  AddImageRenderPart(rw, inline_rp, rp->width, rp->height);
	
	  ci->next = xci->next;
	  xci->next = inline_rp;
	  ci->mp = rp;
	}
	else AddFakeRenderPart(rw);

	return;
      }
    }
  }

  if ((lup = ParseURL(url)) == NULL) return;

  if ((bup = GetBaseURLParts(rw)) != NULL)
  {
    aup = MakeURLParts(lup, bup);
    DestroyURLParts(lup);
  }
  else aup = lup;

  inline_rp = CreateInlineState(rw, p, aup);
  ci = (InlineState *)GetRenderState(inline_rp);

  XtVaGetValues((Widget)rw,
		XtNbackground, &ci->bg.pixel,
		NULL);
  XQueryColor(XtDisplay((Widget)rw), rw->www.cmap, &ci->bg);


  if (GetWidthHeight(p, &width, &height) == 0)
  {
    AddImageRenderPart(rw, inline_rp, width, height);
  }
  else rw->www.delay = True;
  
  url = MakeURL(aup, 0);
  XtCallCallbackList((Widget)rw, rw->www.downloadCallbacks, (XtPointer)url);
  free_mem(url);
  
  aup->reload = rw->www.reload ? 1:0;
  LoadDocument((IOContext)rw, aup, HandleInlineDocument, inline_rp);

  return;
}
