/*
 * layout.c
 *
 * Copyright (c) 1995, John D. Kilburg (john@isri.unlv.edu)
 *
 * See copyright.h for more information.
 *
 */
#include "copyright.h"

#include <stdio.h>
#include <ctype.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Scrollbar.h>

#include "common.h"
#include "WWWP.h"

static void HandleHx _ArgProto((WWWWidget, int, MLTag *, int));
static void HandleP _ArgProto((WWWWidget, MLTag *, int));
static void HandleBR _ArgProto((WWWWidget, MLTag *, int));
static void HandleAddress _ArgProto((WWWWidget, MLTag *, int));
static void HandleA _ArgProto((WWWWidget, MLTag *, int));
static void HandlePRE _ArgProto((WWWWidget, MLTag *, int));
static void HandleListing _ArgProto((WWWWidget, MLTag *, int));
static void HandlePlaintext _ArgProto((WWWWidget, MLTag *, int));
static void HandleBase _ArgProto((WWWWidget, MLTag *, int));
static void HandleBlockquote _ArgProto((WWWWidget, MLTag *, int));
static RenderLine *CreateRenderLine _ArgProto((WWWWidget));

static void ConfigureWidgets _ArgProto((WWWWidget));

static void CallTagHandler _ArgProto((WWWWidget, int, MLTag *));

/*
 * WWWSpacify
 *
 * Turns all whitespace to spaces
 */
void
WWWSpacify(s, len)
char *s;
int len;
{
  char *cp;

  for (cp = s; cp < s + len; cp++)
  {
    if (isspace(*cp)) *cp = ' ';
  }

  return;
}

/*
 * GetBaseURLParts
 */
URLParts *
GetBaseURLParts(rw)
WWWWidget rw;
{
  if (rw->www.bup != NULL) return(rw->www.bup);
  if (rw->www.dilist == NULL) return(NULL);
  return(rw->www.dilist->up);
}

/*
 * GetWidthHeight
 */
int
GetWidthHeight(p, width, height)
MLTag *p;
unsigned int *width, *height;
{
  char *wstr, *hstr;

  wstr = MLFindAttribute(p, "width");
  hstr = MLFindAttribute(p, "height");
  if (wstr != NULL && hstr != NULL)
  {
    *width = (unsigned int)atoi(wstr);
    *height = (unsigned int)atoi(hstr);
    return(0);
  }

  return(-1);
}

void
AddLineBreak(rw)
WWWWidget rw;
{
  CreateRenderLine(rw);

  return;
}

void
AddBlankLine(rw)
WWWWidget rw;
{
  AddLineBreak(rw);
  AddLineBreak(rw);

  return;
}

/*
 * InEnv
 */
Boolean
InEnv(rw, tagid)
WWWWidget rw;
int tagid;
{
  RenderEnv *re;

  for (re = rw->www.restack; re != NULL; re = re->next)
  {
    if (re->tagid == tagid) return(True);
  }

  return(False);
}

/*
 * CreateRenderEnv
 */
RenderEnv *
CreateRenderEnv(rw, tagid)
WWWWidget rw;
int tagid;
{
  RenderEnv *re;

  re = (RenderEnv *)WWWGetMem(rw, sizeof(RenderEnv));
  re->tagid = tagid;

  if (rw->www.restack != NULL)
  {
    RenderEnv *ce = rw->www.restack;

    re->anchorPart = ce->anchorPart;
    re->namePart = ce->namePart;
    re->font = ce->font;
    re->width = ce->width;
    re->align = ce->align;
    re->text_mode = ce->text_mode;
    re->x = ce->x;
    re->y = ce->y;
    re->wrap = ce->wrap;
  }
  else
  {
    re->anchorPart = NULL;
    re->namePart = NULL;
    re->font = rw->www.normalFont;
    re->width = rw->www.width;
    re->align = ALIGN_LEFT;
    re->text_mode = TAG_NOTHING;
    re->x = rw->www.margin;
    re->y = rw->www.margin;
    re->wrap = True;
  }

  re->next = rw->www.restack;
  rw->www.restack = re;

  return(re);
}

/*
 * DestroyRenderEnv2
 */
void
DestroyRenderEnv2(rw, tagid, call_self)
WWWWidget rw;
int tagid;
Boolean call_self;
{
  RenderEnv *fe, *te;
  static Boolean locked = False;

  if (locked) return;
  locked = True;

  for (fe = rw->www.restack; fe != NULL; fe = fe->next)
  {
    if (tagid == fe->tagid) break;
  }

  if (fe != NULL)
  {
    for (; rw->www.restack != NULL && rw->www.restack != fe; )
    {
      te = rw->www.restack;
      CallTagHandler(rw, te->tagid, NULL);
      rw->www.restack = rw->www.restack->next;
    }
    if (call_self) CallTagHandler(rw, fe->tagid, NULL);
    rw->www.restack = fe->next;
  }

  locked = False;

  return;
}

/*
 * DestroyRenderEnv
 */
void
DestroyRenderEnv(rw, tagid)
WWWWidget rw;
int tagid;
{
  DestroyRenderEnv2(rw, tagid, False);
  return;
}

/*
 * DestroyRenderPart
 */
void
DestroyRenderPart(rp)
RenderPart *rp;
{
  if (rp->destroy != NULL) (rp->destroy)(rp->state);

  return;
}

/*
 * CreateRenderPart
 */
RenderPart *
CreateRenderPart(rw, p, setup, render, destroy, event, state_size)
WWWWidget rw;
MLTag *p;
void (*setup)();
void (*render)();
void (*destroy)();
void (*event)();
int state_size;
{
  RenderPart *rp;

  rp = (RenderPart *)WWWGetMem(rw, sizeof(RenderPart));
  rp->p = p;
  rp->anchorPart = rw->www.restack->anchorPart;
  rp->namePart = rw->www.restack->namePart;

  if (rp->anchorPart != NULL)
  {
    if (rw->www.aphead == NULL) rw->www.aphead = rp;
    else rw->www.aptail->anext = rp;
    rw->www.aptail = rp;
  }

  rp->setup = setup;
  rp->render = render;
  rp->destroy = destroy;
  rp->event = event;
  if (state_size > 0) rp->state = (RenderPart *)WWWGetMem(rw, state_size);
  else rp->state = NULL;

  rp->x = 0;
  rp->y = 0;
  rp->width = 0;
  rp->height = 0;
  rp->renderable = False;
  rp->lp = NULL;
  rp->lnext = NULL;
  rp->mnext = NULL;
  rp->anext = NULL;

  /*
   * Add to the master list.
   */
  if (rw->www.rphead == NULL) rw->www.rphead = rp;
  else rw->www.rptail->mnext = rp;
  rw->www.rptail = rp;

  return(rp);
}

/*
 * AddRenderPart
 *
 */
int
AddRenderPart(rw, rp, width, height, longline)
WWWWidget rw;
RenderPart *rp;
unsigned int width, height;
Boolean longline;
{
  RenderLine *rl = rw->www.rltail;

  rp->width = width;
  rp->height = height;

  /*
   * If there are parts in the line and it will go off the end of the
   * line then create a new line.
   *
   * If a part is the only part on a line then it is never rejected.
   */
  if (rw->www.restack->wrap && rl->rphead != NULL && !longline &&
      rl->x + rp->width > rw->www.restack->x + rw->www.restack->width)
  {
    rl = CreateRenderLine(rw);
  }

  rp->x = rl->x;
  rl->x += rp->width;

  rp->lp = rl;

  /*
   * Add to the line list.
   */
  if (rl->rptail != NULL) rl->rptail->lnext = rp;
  else rl->rphead = rp;
  rl->rptail = rp;

  rl->miny = 0;
  if (rp->height > rl->maxy) rl->maxy = rp->height;

  if (rw->www.delay)
  {
    rw->www.delay = False;
    WWWLayout(rw, NULL);
  }

  return(0);
}

/*
 * CreateRenderLine
 */
static RenderLine *
CreateRenderLine(rw)
WWWWidget rw;
{
  RenderLine *rl;
  RenderPart *rp;
  int adjustx;
  Region r;
  Position x, y;
  Dimension width, height;
  XRectangle rec;
  unsigned int wm;

  rl = (RenderLine *)WWWGetMem(rw, sizeof(RenderLine));
  rl->rphead = NULL;
  rl->rptail = NULL;
  rl->height = 0;
  rl->next = NULL;
  rl->x = rw->www.restack->x;
  rl->miny = 0;
  rl->maxy = 0;
  rl->align = rw->www.restack->align;

  if (rw->www.rltail != NULL)
  {
    RenderLine *ll = rw->www.rltail;

    ll->next = rl;
    wm = (unsigned int)ll->x + (unsigned int)rw->www.margin;
    if (wm > rw->www.width)
    {
      rw->www.width = wm;
      adjustx = 0;
    }
    else
    {
      if (ll->align == ALIGN_CENTER) adjustx = (rw->www.width - ll->x) / 2;
      else if (ll->align == ALIGN_RIGHT) adjustx = rw->www.width - ll->x;
      else adjustx = 0;
    }

    ll->height = ll->maxy - ll->miny;
    if (ll->miny < 0) ll->y -= ll->miny;

    rl->y = rw->www.lineSpace + ll->y + ll->height;
    rw->www.height = ll->y + ll->height + rw->www.margin;

    /*
     * Search for the region that is on the screen
     */
    r = XCreateRegion();
    XtVaGetValues(rw->www.child, XtNx, &x, XtNy, &y, NULL);
    XtVaGetValues(rw->www.clip, XtNwidth, &width, XtNheight, &height, NULL);
    rec.x = (short)x;
    rec.y = (short)x;
    rec.width = (unsigned short)width;
    rec.height = (unsigned short)height;
    XUnionRectWithRegion(&rec, r, r);

    for (rp = ll->rphead; rp != NULL; rp = rp->lnext)
    {
      rp->x += adjustx;
      rp->renderable = True;

      if (rp->setup != NULL) (rp->setup)(rw, rp, r, ll->y, ll->height);
    }

    XDestroyRegion(r);
  }
  else
  {
    rw->www.rlhead = rl;
    rl->y = rw->www.margin;
  }

  rw->www.rltail = rl;

  ConfigureWidgets(rw);

  return(rl);
}

static void
HandleA(rw, p, end)
WWWWidget rw;
MLTag *p;
int end;
{
  if (end != 0) DestroyRenderEnv(rw, TAG_A);
  else
  {
    RenderEnv *re;

    re = CreateRenderEnv(rw, TAG_A);

    if (MLFindAttribute(p, "href") != NULL) re->anchorPart = p;
    if (MLFindAttribute(p, "name") != NULL) re->namePart = p;
  }

  return;
}

static void
HandleHx(rw, tag, p, end)
WWWWidget rw;
int tag;
MLTag *p;
int end;
{
  RenderEnv *re;

  if (end != 0)
  {
    AddBlankLine(rw);
    DestroyRenderEnv(rw, tag);
    AddBlankLine(rw);
  }
  else
  {
    char *align;

    AddBlankLine(rw);

    re = CreateRenderEnv(rw, tag);
    if (p->id == TAG_H1) re->font = rw->www.h1Font;
    else if (p->id == TAG_H2) re->font = rw->www.h2Font;
    else if (p->id == TAG_H3) re->font = rw->www.h3Font;
    else if (p->id == TAG_H4) re->font = rw->www.h4Font;
    else if (p->id == TAG_H5) re->font = rw->www.h5Font;
    else if (p->id == TAG_H6) re->font = rw->www.h6Font;

    if ((align = MLFindAttribute(p, "align")) != NULL)
    {
      re->align = MLGetAlignID(align);
    }

    AddBlankLine(rw);
  }
  
  return;
}

static void
HandleFontChange(rw, end, tag)
WWWWidget rw;
int end;
int tag;
{
  RenderEnv *re;

  if (end != 0) DestroyRenderEnv(rw, tag);
  else
  {
    re = CreateRenderEnv(rw, tag);
    if (tag == TAG_EM) re->font = rw->www.iFont;
    else if (tag == TAG_B) re->font = rw->www.bFont;
    else if (tag == TAG_STRONG) re->font = rw->www.bFont;
    else if (tag == TAG_TT) re->font = rw->www.ttFont;
    else if (tag == TAG_CITE) re->font = rw->www.iFont;
    else if (tag == TAG_I) re->font = rw->www.iFont;
  }
  
  return;
}

static void
HandleTitle(rw, end)
WWWWidget rw;
int end;
{
  if (end != 0) DestroyRenderEnv(rw, TAG_TITLE);
  else
  {
    RenderEnv *re;

    re = CreateRenderEnv(rw, TAG_TITLE);
    re->text_mode = TAG_TITLE;
  }

  return;
}

static void
HandleP(rw, p, end)
WWWWidget rw;
MLTag *p;
int end;
{
  RenderEnv *re;

  if (end != 0)
  {
    DestroyRenderEnv(rw, TAG_P);
    AddBlankLine(rw);
  }
  else
  {
    char *align;

    AddBlankLine(rw);
    re = CreateRenderEnv(rw, TAG_P);
    if ((align = MLFindAttribute(p, "align")) != NULL)
    {
      re->align = MLGetAlignID(align);
    }
    AddBlankLine(rw);
  }

  return;
}

static void
HandleBR(rw, p, end)
WWWWidget rw;
MLTag *p;
int end;
{
  if (end == 0)
  {
    char *align;
    RenderEnv *re;

    re = CreateRenderEnv(rw, TAG_BR);

    if ((align = MLFindAttribute(p, "align")) != NULL)
    {
      re->align = MLGetAlignID(align);
    }

    AddLineBreak(rw);

    DestroyRenderEnv(rw, TAG_BR);
  }

  return;
}

static void
HandleAddress(rw, p, end)
WWWWidget rw;
MLTag *p;
int end;
{
  if (end != 0)
  {
    DestroyRenderEnv(rw, TAG_ADDRESS);
    AddLineBreak(rw);
  }
  else
  {
    RenderEnv *re;

    AddLineBreak(rw);
    re = CreateRenderEnv(rw, TAG_ADDRESS);
    re->font = rw->www.iFont;
  }

  return;
}

static void
HandlePRE(rw, p, end)
WWWWidget rw;
MLTag *p;
int end;
{
  if (end != 0)
  {
    DestroyRenderEnv(rw, TAG_PRE);
    AddBlankLine(rw);
  }
  else
  {
    RenderEnv *re;

    re = CreateRenderEnv(rw, TAG_PRE);
    re->font = rw->www.ttFont;
    re->width = 10000000;
    re->text_mode = TAG_PRE;
    re->wrap = False;
    AddBlankLine(rw);
  }

  return;
}

static void
HandleListing(rw, p, end)
WWWWidget rw;
MLTag *p;
int end;
{
  if (end != 0)
  {
    AddBlankLine(rw);
    DestroyRenderEnv(rw, TAG_LISTING);
    AddBlankLine(rw);
  }
  else
  {
    RenderEnv *re;

    AddBlankLine(rw);
    re = CreateRenderEnv(rw, TAG_LISTING);
    re->font = rw->www.ttFont;
    re->width = 10000000;
    re->text_mode = TAG_LISTING;
    AddBlankLine(rw);
  }

  return;
}

static void
HandlePlaintext(rw, p, end)
WWWWidget rw;
MLTag *p;
int end;
{
  if (end != 0)
  {
    AddBlankLine(rw);
    DestroyRenderEnv(rw, TAG_PLAINTEXT);
    AddBlankLine(rw);
  }
  else
  {
    RenderEnv *re;

    AddBlankLine(rw);
    re = CreateRenderEnv(rw, TAG_PLAINTEXT);
    re->font = rw->www.ttFont;
    re->width = 10000000;
    re->wrap = False;
    AddBlankLine(rw);
  }

  return;
}

static void
HandleBase(rw, p, end)
WWWWidget rw;
MLTag *p;
int end;
{
  char *url;
  URLParts *up;

  if (end == 0)
  {
    if ((url = MLFindAttribute(p, "href")) != NULL)
    {
      if ((up = ParseURL(url)) != NULL)
      {
	if (rw->www.bup != NULL) DestroyURLParts(rw->www.bup);
	rw->www.bup = up;
      }
    }
  }

  return;
}

static void
HandleMisc(rw, tagid, p, end)
WWWWidget rw;
int tagid;
MLTag *p;
int end;
{
  if (end == 0) CreateRenderEnv(rw, tagid);
  else DestroyRenderEnv(rw, tagid);

  return;
}

static void
HandleBlockquote(rw, p, end)
WWWWidget rw;
MLTag *p;
int end;
{
  RenderEnv *re;

  if (end != 0)
  {
    AddBlankLine(rw);
    DestroyRenderEnv(rw, TAG_BLOCKQUOTE);
    AddBlankLine(rw);
  }
  else
  {
    AddBlankLine(rw);
    re = CreateRenderEnv(rw, TAG_BLOCKQUOTE);
    re->align = ALIGN_CENTER;
    re->width = re->width - re->width * 20 / 100;
    re->font = rw->www.iFont;
    AddBlankLine(rw);
  }

  return;
}

/*
 * ConfigureWidgets
 *
 * One day the second argument needs to be removed because there will
 * be multiple layouts with all sorts of weird regions that will have to
 * calculated to determine the size of the document.
 */
static void
ConfigureWidgets(rw)
WWWWidget rw;
{
  Dimension child_width, child_height;
  Dimension clip_width, clip_height;
  Dimension vx, hy;
  float shown;

  if (!XtIsRealized((Widget)rw)) return;

  /*
   * Big fun.  Why can't dimensions be at least 32 bits long?
   */
  if (rw->www.width > 0x7ffd) child_width = 0x7ffd;
  else child_width = (Dimension)rw->www.width;

  if (rw->www.height > 0x7ffd) child_height = 0x7ffd;
  else child_height = (Dimension)rw->www.height;

  if (child_height > rw->www.clip->core.height) vx = rw->www.vert_bar_width;
  else
  {
    vx = 0;
    child_height = rw->www.clip->core.height;
  }

  if (child_width > rw->www.clip->core.width) hy = rw->www.horiz_bar_height;
  else hy = 0;

  clip_width = rw->core.width - rw->www.vert_bar_width;
  clip_height = rw->core.height;

  if (child_height > clip_height)
  {
    clip_height -= hy;

    XtConfigureWidget(rw->www.vert_bar,
		      0, hy,
		      rw->www.vert_bar_width, rw->core.height - hy, 0);
    shown = (float)clip_height / (float)child_height;
    XawScrollbarSetThumb(rw->www.vert_bar, (float)-1.0, shown);
    if (!XtIsManaged(rw->www.vert_bar)) XtManageChild(rw->www.vert_bar);
  }
  else XtUnmanageChild(rw->www.vert_bar);

  if (child_width > clip_width)
  {
    XtConfigureWidget(rw->www.horiz_bar,
		      vx, 0,
		      rw->core.width - vx, rw->www.horiz_bar_height, 0);
    shown = (float)clip_width / (float)child_width;
    XawScrollbarSetThumb(rw->www.horiz_bar, (float)-1.0, shown);
    if (!XtIsManaged(rw->www.horiz_bar)) XtManageChild(rw->www.horiz_bar);
  }
  else XtUnmanageChild(rw->www.horiz_bar);

  XtConfigureWidget(rw->www.clip,
		    rw->www.vert_bar_width, hy,
		    clip_width, clip_height, 0);

  XtResizeWidget(rw->www.child, child_width, child_height, 0);

  return;
}

/*
 * CallTagHandler
 */
static void
CallTagHandler(rw, tag, p)
WWWWidget rw;
int tag;
MLTag *p;
{
  int end;

  if (p == NULL) end = 1;
  else end = p->end;

  if (tag == TAG_EOF) CreateRenderLine(rw);
  else if (tag == TAG_H1) HandleHx(rw, tag, p, end);
  else if (tag == TAG_H2) HandleHx(rw, tag, p, end);
  else if (tag == TAG_H3) HandleHx(rw, tag, p, end);
  else if (tag == TAG_H4) HandleHx(rw, tag, p, end);
  else if (tag == TAG_H5) HandleHx(rw, tag, p, end);
  else if (tag == TAG_H6) HandleHx(rw, tag, p, end);
  else if (tag == TAG_IMG) HandleImg(rw, p, end);
  else if (tag == TAG_TITLE) HandleTitle(rw, end);
  else if (tag == TAG_HR) HandleHR(rw, p, end);
  else if (tag == TAG_P) HandleP(rw, p, end);
  else if (tag == TAG_UL) HandleList(rw, tag, p, end);
  else if (tag == TAG_OL) HandleList(rw, tag, p, end);
  else if (tag == TAG_MENU) HandleList(rw, tag, p, end);
  else if (tag == TAG_DIR) HandleList(rw, tag, p, end);
  else if (tag == TAG_DL) HandleList(rw, tag, p, end);
  else if (tag == TAG_LI) HandleLI(rw, p, end);
  else if (tag == TAG_DT) HandleDT(rw, p, end);
  else if (tag == TAG_DD) HandleDD(rw, p, end);
  else if (tag == TAG_OL) HandleOL(rw, p, end);
  else if (tag == TAG_BR) HandleBR(rw, p, end);
  else if (tag == TAG_ADDRESS) HandleAddress(rw, p, end);
  else if (tag == TAG_A) HandleA(rw, p, end);
  else if (tag == TAG_PRE) HandlePRE(rw, p, end);
  else if (tag == TAG_LISTING) HandleListing(rw, p, end);
  else if (tag == TAG_PLAINTEXT) HandlePlaintext(rw, p, end);
  else if (tag == TAG_B) HandleFontChange(rw, end, tag);
  else if (tag == TAG_EM) HandleFontChange(rw, end, tag);
  else if (tag == TAG_STRONG) HandleFontChange(rw, end, tag);
  else if (tag == TAG_TT) HandleFontChange(rw, end, tag);
  else if (tag == TAG_CITE) HandleFontChange(rw, end, tag);
  else if (tag == TAG_I) HandleFontChange(rw, end, tag);
  else if (tag == TAG_KBD) HandleFontChange(rw, end, tag);
  else if (tag == TAG_FORM) HandleForm(rw, p, end);
  else if (tag == TAG_INPUT) HandleInput(rw, p, end);
  else if (tag == TAG_TEXTAREA) HandleTextarea(rw, p, end);
  else if (tag == TAG_BASE) HandleBase(rw, p, end);
  else if (tag == TAG_HEAD) HandleMisc(rw, tag, p, end);
  else if (tag == TAG_HTML) HandleMisc(rw, tag, p, end);
  else if (tag == TAG_BODY) HandleMisc(rw, tag, p, end);
  else if (tag == TAG_BLOCKQUOTE) HandleBlockquote(rw, p, end);
  else if (tag == TAG_OPTION) HandleOption(rw, p, end);
  else if (tag == TAG_SELECT) HandleSelect(rw, p, end);

  return;
}

/*
 * WWWLayout
 */
void
WWWLayout(rw, p)
WWWWidget rw;
MLTag *p;
{
  if (p != NULL)
  {
    if (rw->www.pqhead == NULL) rw->www.pqhead = p;
    else rw->www.pqtail->qnext = p;
    rw->www.pqtail = p;
  }

  for (; !rw->www.delay && rw->www.pqhead != NULL;
       rw->www.pqhead = rw->www.pqhead->qnext)
  {
    p = rw->www.pqhead;

    if (rw->www.restack->tagid == TAG_PLAINTEXT)
    {
      PlainText(rw, p, True);
    }
    else if (p->id == TAG_TEXT)
    {
      switch (rw->www.restack->text_mode)
      {
	case TAG_TEXTAREA:
          TextareaText(rw, p);
	  break;
	case TAG_OPTION:
          OptionText(rw, p);
	  break;
	case TAG_AU:
	  AuthorText(rw, p);
	  break;
	case TAG_TITLE:
          TitleText(rw, p);
	  break;
        case TAG_PRE:
	case TAG_LISTING:
	case TAG_CODE:
	  PlainText(rw, p, False);
	  break;
        default:
	  FillText(rw, p);
      }
    }
    else CallTagHandler(rw, p->id, p);
  }

  return;
}

/*
 * WWWEndLayout
 */
void
WWWEndLayout(rw)
WWWWidget rw;
{
  DestroyRenderEnv(rw, TAG_START);

  if (rw->www.rphead != NULL)
  {
    RenderPart *t, *c;
    
    for (c = rw->www.rphead; c != NULL; )
    {
      t = c;
      c = c->mnext;
      DestroyRenderPart(t);
    }
  }

  if (rw->www.title != NULL) free_mem(rw->www.title);
  if (rw->www.author != NULL) free_mem(rw->www.author);
  if (rw->www.bup != NULL) DestroyURLParts(rw->www.bup);
  if (rw->www.hs != NULL) MLDestroy(rw->www.hs);
  if (rw->www.memstate != NULL) WWWFreeMem(rw);

  WWWInitLayoutVars(rw);

  return;
}

/*
 * InitLayoutVars
 */
void
WWWInitLayoutVars(rw)
WWWWidget rw;
{
  rw->www.aphead = NULL;
  rw->www.aptail = NULL;
  rw->www.rphead = NULL;
  rw->www.rptail = NULL;
  rw->www.rlhead = NULL;
  rw->www.rltail = NULL;
  rw->www.restack = NULL;
  rw->www.pqhead = NULL;
  rw->www.pqtail = NULL;
  rw->www.delay = False;
  rw->www.bup = NULL;
  rw->www.title = NULL;
  rw->www.author = NULL;
  rw->www.formstate = NULL;
  rw->www.memstate = NULL;
  rw->www.hs = NULL;

  return;
}

/*
 * WWWStartLayout
 */
void
WWWStartLayout(rw, up, raw)
WWWWidget rw;
URLParts *up;
Boolean raw;
{
  RenderEnv *re;
  Dimension clip_width, clip_height;

  WWWInitLayoutVars(rw);

  /*
   * Do nothing if the window is too narrow.
   */
  if (rw->www.vert_bar_width > rw->core.width + 100) return;
  if (rw->www.horiz_bar_height > rw->core.height + 100) return;

  rw->www.hs = MLInit(rw, WWWLayout);

  /*
   * Set the initial dimensions of the clip widget (they don't change)
   * and the initial size of the document.
   */
  clip_width = rw->core.width - rw->www.vert_bar_width;
  clip_height = rw->core.height;

  rw->www.width = (unsigned int)(clip_width - rw->www.margin * 2);
  rw->www.owidth = rw->www.width;
  rw->www.height = 0;

  if (XtIsRealized(rw->www.child))
  {
    XClearWindow(rw->www.dpy, XtWindow(rw->www.child));

    XtConfigureWidget(rw->www.clip,
		      rw->www.vert_bar_width, 0,
		      clip_width, clip_height,
		      0);

    XtConfigureWidget(rw->www.child,
		      0, 0,
		      clip_width, clip_height, 0);

    ConfigureWidgets(rw);
  }

  /*
   * Setup environment
   */
  re = CreateRenderEnv(rw, TAG_START);
  re->x = rw->www.margin;
  re->y = rw->www.margin;
  re->width = rw->www.width;
  re->font = rw->www.normalFont;

  if (raw)
  {
    re = CreateRenderEnv(rw, TAG_PLAINTEXT);
    re->font = rw->www.ttFont;
    re->width = 10000000;
  }

  /*
   * Intialize line queue
   */
  CreateRenderLine(rw);

  return;
}

