/*
 * convert.c
 *
 * Copyright (C) 1993, 1994, 1995, John D. Kilburg (john@cs.unlv.edu)
 *
 * See copyright.h for details.
 */
#include "copyright.h"

#include <stdio.h>
#include <sys/param.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "common.h"
#include "wwwio.h"
#include "wwwioextern.h"

#ifndef MAXPATHLEN
#define MAXPATHLEN 1024
#endif

typedef struct _convert
{
  char *use;
  char *incontent;
  char *outcontent;
  char *command;
  int (*func)();
  struct _convert *next;
} Convert;

typedef struct _convert_info
{
  Document *d;
  int fd[2];
  void (*lfunc)();
  int (*func)();
  void *closure;
  Convert *c;
  IOContext ioc;
} ConvertInfo;

static int PlainTextToHTML _ArgProto((char *, int, char **, int *));

static Convert converters[] =
{
  { "*", "text/plain", "text/html", NULL, PlainTextToHTML },
  { NULL, NULL, NULL, NULL, NULL },
};

static Convert *clist = NULL;

/*
 * ReadConvertDB
 *
 * Reads in the convert entries in a list of files separated by colons.
 * For example,
 *
 * ~/.chimera_convert:~john/lib/convert:/local/infosys/lib/convert
 */
void
ReadConvertFiles(filelist)
char *filelist;
{
  Convert *t = NULL, *c;
  char *f;
  char *filename;
  char buffer[BUFSIZ];
  char use[BUFSIZ];
  char incontent[BUFSIZ];
  char outcontent[BUFSIZ];
  char command[BUFSIZ];
  FILE *fp;

  f = filelist;
  while ((filename = mystrtok(f, ':', &f)) != NULL)
  {
    filename = FixFilename(filename);
    if (filename == NULL) continue;

    fp = fopen(filename, "r");
    if (fp == NULL) continue;
    
    while (fgets(buffer, sizeof(buffer), fp) != NULL)
    {
      if (buffer[0] == '#' || buffer[0] == '\n') continue;

      if (sscanf(buffer, "%s %s %s %[^\n]",
		 use, incontent, outcontent, command) == 4)
      {
	c = (Convert *)alloc_mem(sizeof(Convert));
	c->use = alloc_string(use);
	c->incontent = alloc_string(incontent);
	c->outcontent = alloc_string(outcontent);
	c->command = strcasecmp(command, "none") == 0 ?
	    NULL:alloc_string(command);
	c->func = NULL;
	c->next = NULL;
	
	if (clist) t->next = c;
	else clist = c;
	t = c;
      }
    }
    
    fclose(fp);
  }

  return;
}

/*
 * GetConvert
 *
 * Returns an entry from the content file that matches the access method
 * and content-type.
 */
static Convert *
GetConvert(use, incontent)
char *use;
char *incontent;
{
  Convert *c;
  int flen, elen;
  int i;

  flen = strlen(incontent);
  c = clist;
  while (c)
  {
    if (c->use[0] == '*' || strcasecmp(use, c->use) == 0)
    {
      if (c->incontent[0] == '*')
      {
	return(c);
      }
      else
      {
	elen = strlen(c->incontent);
	if (elen <= flen && strncasecmp(c->incontent, incontent, elen) == 0)
	{
	  return(c);
	}
      }
    }
    c = c->next;
  }

  for (i = 0; converters[i].use != NULL; i++)
  {
    if (converters[i].use[0] == '*' || strcasecmp(use, converters[i].use) == 0)
    {
      if (converters[i].incontent[0] == '*')
      {
	return(converters + i);
      }
      else
      {
	elen = strlen(converters[i].incontent);
	if (elen <= flen &&
	    strncasecmp(converters[i].incontent, incontent, elen) == 0)
	{
	  return(converters + i);
	}
      }
    }
  }

  return(NULL);
}

/*
 * PlainTextToHTML
 */
static int
PlainTextToHTML(t, tlen, h, hlen)
char *t;
int tlen;
char **h;
int *hlen;
{
  *hlen = tlen + 12;
  *h = alloc_mem(*hlen);
  strcpy(*h, "<plaintext>");
  strcat(*h, t);

  return(0);
}

/*
 * convert_destroy_info
 */
static void
convert_destroy_info(ci)
ConvertInfo *ci;
{
  free_mem((char *)ci);

  return;
}

/*
 * convert_clean
 */
static void
convert_clean(s, closure, error)
int s;
char *closure;
int error;
{
  ConvertInfo *ci = (ConvertInfo *)closure;

  IOEnd(ci->ioc, s);
  close(ci->fd[0]);

  convert_destroy_info(ci);

  return;
}

/*
 * convert_data
 */
static void
convert_data(s, t, tlen, closure)
int s;
char *t;
int tlen;
char *closure;
{
  ConvertInfo *ci = (ConvertInfo *)closure;

  IOEnd(ci->ioc, s);
  close(ci->fd[0]);

  (ci->lfunc)(t, tlen, ci);

  return;
}

/*
 * SetupConverter
 */
static void
SetupConverter(ioc, d, c, path, lfunc, func, closure)
IOContext ioc;
Document *d;
Convert *c;
char *path;
void (*lfunc)();
int (*func)();
void *closure;
{
  char *cmdline;
  char filename[MAXPATHLEN + 1];
  char *data;
  int datalen;
  ConvertInfo *ci;

  if (c->func != NULL)
  {
    if ((c->func)(d->text, d->len, &data, &datalen) != 0)
    {
      ; /* convfail */
    }
    else
    {
      d->ptext = data;
      d->plen = datalen;
      AddMIMEField(d, "x-pcontent-type", c->outcontent);
      (func)(d, closure);
    }
  }
  else
  {
    if (SaveData(d->text, d->len, tmpnam(filename)) == -1)
    {
      ; /* nosave */
    }
    else
    {
      ci = (ConvertInfo *)alloc_mem(sizeof(ConvertInfo));
      ci->d = d;
      ci->func = func;
      ci->closure = closure;
      ci->lfunc = lfunc;
      ci->c = c;
      ci->ioc = ioc;

      cmdline = CreateCommandLine(c->command, path, filename);

      if (PipeCommand(cmdline, ci->fd, 1) != 0)
      {
	; /* pipe failed */
      }
      else
      {
        IOStart(ioc, ci->fd[1], convert_clean, ci);
	IOReadStream(ioc, ci->fd[1], 0, convert_data, 1);
      }

      free_mem(cmdline);
    }
  }

  return;
}

/*
 * TransferEncoding
 */
static void
TransferEncoding(t, tlen, ci)
char *t;
int tlen;
ConvertInfo *ci;
{
  return;
}

/*
 * Encoding
 */
static void
Encoding(t, tlen, ci)
char *t;
int tlen;
ConvertInfo *ci;
{
  free_mem(ci->d->text);
  ci->d->text = t;
  ci->d->len = tlen;

  (ci->func)(ci->d, ci->closure);

  convert_destroy_info(ci);

  return;
}

/*
 * Regular
 */
static void
Regular(t, tlen, ci)
char *t;
int tlen;
ConvertInfo *ci;
{
  ci->d->ptext = t;
  ci->d->plen = tlen;
  AddMIMEField(ci->d, "x-pcontent-type", ci->c->outcontent);

  (ci->func)(ci->d, ci->closure);

  convert_destroy_info(ci);

  return;
}

/*
 * ConvertDocument
 *
 * Converts document according to instructions in the 'content file'.
 * The content-type is used to identify the type of file and pick the
 * the correct entry in the content file.  The access method is also
 * used (http, ftp, inline, ...)
 */
void
ConvertDocument(ioc, use, path, d, func, closure)
IOContext ioc;
char *use;
char *path;
Document *d;
int (*func)();
void *closure;
{
  char *value;
  char *content;
  Convert *c;

  if ((content = GetMIMEValue(d, "content-type", 0)) == NULL)
  {
    content = "application/octet";
  }

  if ((value = GetMIMEValue(d, "content-transfer-encoding", 1)) != NULL)
  {
    if ((c = GetConvert(value, content)) != NULL)
    {
      SetupConverter(ioc, d, c, path, TransferEncoding, func, closure);
    }
  }
  else if ((value = GetMIMEValue(d, "content-encoding", 1)) != NULL)
  {
    if ((c = GetConvert(value, content)) != NULL)
    {
      SetupConverter(ioc, d, c, path, Encoding, func, closure);
    }
  }
  else
  {
    if (use == NULL) use = "none";
    if ((c = GetConvert(use, content)) != NULL)
    {
      SetupConverter(ioc, d, c, path, Regular, func, closure);
    }
  }

  return;
}
